package edu.berkeley.qfat;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import java.util.*;
import edu.berkeley.qfat.geom.*;
import edu.berkeley.qfat.geom.Point;

// FIXME: re-orient goal (how?)

public class Main extends MeshViewer {

    public static final Random random = new Random();
    
    /** magnification factor */
    private static final float MAG = 1;

    public Main(StlFile stlf, Frame f) {
        super(f);

        for(int i=0; i<stlf.coordArray.length; i+=3) {
            Point p0 = new Point(stlf.coordArray[i+0].x * MAG, stlf.coordArray[i+0].y * MAG, stlf.coordArray[i+0].z * MAG);
            Point p1 = new Point(stlf.coordArray[i+1].x * MAG, stlf.coordArray[i+1].y * MAG, stlf.coordArray[i+1].z * MAG);
            Point p2 = new Point(stlf.coordArray[i+2].x * MAG, stlf.coordArray[i+2].y * MAG, stlf.coordArray[i+2].z * MAG);
            Vec n  = new Vec(stlf.normArray[i/3].x * MAG, stlf.normArray[i/3].y  * MAG, stlf.normArray[i/3].z * MAG);
            Mesh.T t  = goal.newT(p0, p1, p2, n);
        }

        // rotate to align major axis -- this probably needs to be done by a human.
        goal.transform(new Matrix(new Vec(0, 0, 1), (float)(Math.PI/2)));

        float goal_width  = goal.diagonal().dot(new Vec(1, 0, 0));
        float goal_height = goal.diagonal().dot(new Vec(0, 1, 0));
        float goal_depth  = goal.diagonal().dot(new Vec(0, 0, 1));

        float width  = (float)0.6;
        float height = (float)0.08;
        float depth  = (float)0.3;
        translations = new Matrix[] {

            new Matrix(new Vec(-(width/2),  height,    0)),
            new Matrix(new Vec( (width/2),  height,    0)),
            new Matrix(new Vec(-(width/2), -height,    0)),
            new Matrix(new Vec( (width/2), -height,    0)),
            new Matrix(new Vec(-(width/2),       0,  depth)),
            new Matrix(new Vec( (width/2),       0,  depth)),
            new Matrix(new Vec(-(width/2),       0, -depth)),
            new Matrix(new Vec( (width/2),       0, -depth)),

            new Matrix(new Vec( width,           0,    0)),
            new Matrix(new Vec(-width,           0,    0)),
            /*
            new Matrix(new Vec(     0,           0,    depth)),
            new Matrix(new Vec(     0,           0,   -depth)),
            */
        };


        Point ltf = new Point(-(width/2),  (height/2),  (depth/2));
        Point mtf = new Point( 0.0,        (height/2),  (depth/2));
        Point rtf = new Point( (width/2),  (height/2),  (depth/2));
        Point ltn = new Point(-(width/2),  (height/2), -(depth/2));
        Point mtn = new Point( 0.0,        (height/2), -(depth/2));
        Point rtn = new Point( (width/2),  (height/2), -(depth/2));
        Point lbf = new Point(-(width/2), -(height/2),  (depth/2));
        Point mbf = new Point( 0.0,       -(height/2),  (depth/2));
        Point rbf = new Point( (width/2), -(height/2),  (depth/2));
        Point lbn = new Point(-(width/2), -(height/2), -(depth/2));
        Point mbn = new Point( 0.0,       -(height/2), -(depth/2));
        Point rbn = new Point( (width/2), -(height/2), -(depth/2));
        
        Point[] points = new Point[] {
            ltf,
            mtf,
            rtf,
            ltn,
            mtn,
            rtn,
            lbf,
            mbf,
            rbf,
            lbn,
            mbn,
            rbn
        };


        // top
        tile.newT(ltf, mtf, mtn, null);
        tile.newT(mtn, ltn, ltf, null);
        tile.newT(mtf, rtf, rtn, null);
        tile.newT(rtn, mtn, mtf, null);

        // bottom (swap normals)
        tile.newT(mbf, lbf, mbn, null);
        tile.newT(lbn, mbn, lbf, null);
        tile.newT(rbf, mbf, rbn, null);
        tile.newT(mbn, rbn, mbf, null);
        
        // left
        tile.newT(ltf, ltn, lbn, null);
        tile.newT(lbn, lbf, ltf, null);

        // right (swap normals)
        tile.newT(rtn, rtf, rbn, null);
        tile.newT(rbf, rbn, rtf, null);

        // front
        tile.newT(ltn, mtn, mbn, null);
        tile.newT(ltn, mbn, lbn, null);
        tile.newT(mtn, rtn, rbn, null);
        tile.newT(mtn, rbn, mbn, null);

        // back
        tile.newT(mtf, ltf, mbf, null);
        tile.newT(mbf, ltf, lbf, null);
        tile.newT(rtf, mtf, rbf, null);
        tile.newT(rbf, mtf, mbf, null);

        for(Matrix m : translations) {
            for(Mesh.T t1 : tile) {
                for(Mesh.T t2 : tile) {
                    if (t1==t2) continue;

                    if ((t1.v1().p.times(m).minus(t2.v1().p).mag() < Mesh.EPSILON) &&
                        (t1.v2().p.times(m).minus(t2.v3().p).mag() < Mesh.EPSILON) &&
                        (t1.v3().p.times(m).minus(t2.v2().p).mag() < Mesh.EPSILON)) {
                        t1.e1().bind(t2.e3().pair);
                        t1.e2().bind(t2.e2().pair);
                        t1.e3().bind(t2.e1().pair);
                    }
                    if ((t1.v2().p.times(m).minus(t2.v1().p).mag() < Mesh.EPSILON) &&
                        (t1.v3().p.times(m).minus(t2.v3().p).mag() < Mesh.EPSILON) &&
                        (t1.v1().p.times(m).minus(t2.v2().p).mag() < Mesh.EPSILON)) {
                        t1.e2().bind(t2.e3().pair);
                        t1.e3().bind(t2.e2().pair);
                        t1.e1().bind(t2.e1().pair);
                    }
                    if ((t1.v3().p.times(m).minus(t2.v1().p).mag() < Mesh.EPSILON) &&
                        (t1.v1().p.times(m).minus(t2.v3().p).mag() < Mesh.EPSILON) &&
                        (t1.v2().p.times(m).minus(t2.v2().p).mag() < Mesh.EPSILON)) {
                        t1.e3().bind(t2.e3().pair);
                        t1.e1().bind(t2.e2().pair);
                        t1.e2().bind(t2.e1().pair);
                    }
                }
            }
        }

        //xMesh.Vert mid = lbf.getE(mbn).shatter();

        // rescale to match volume
        float factor = (float)Math.pow(tile.volume() / goal.volume(), 1.0/3.0);
        goal.transform(new Matrix(factor));

        // translate to match centroid
        goal.transform(new Matrix(tile.centroid().minus(goal.centroid())));

        //tx.e2.shatter();
        //tx.e3.shatter();


        tile.rebindPoints();

        //mid.move(new Vec((float)0,0,(float)-0.05));
        //ltn.move(new Vec((float)0,0,(float)-0.05));

        //mtf.move(new Vec(0, (float)-0.05, (float)0.05));


        System.out.println("tile volume: " + tile.volume());
        System.out.println("goal volume: " + goal.volume());

        tile.score_against = goal;
        goal.score_against = tile;
    }

    public synchronized void breakit() {
        if (verts > 200) return;
        PriorityQueue<Mesh.E> es = new PriorityQueue<Mesh.E>();
        for(Mesh.E e : tile.edges()) es.add(e);
        for(int i=0; i<10; i++) {
            Mesh.E e = es.poll();
            verts++;
            System.out.println("shatter " + e);
            e.shatter();
            tile.rebindPoints();
        }
    }

    public synchronized void rand(double temperature, Mesh.Vert p) {
        double tile_score = tile.score();
        double goal_score = goal.score();
        p.rescore();

        Vec v = new Vec((random.nextFloat() - (float)0.5) / 1000,
                        (random.nextFloat() - (float)0.5) / 1000,
                        (random.nextFloat() - (float)0.5) / 1000);
        //Matrix inv = p.errorQuadric();
        //Vec v = new Vec(inv.d, inv.h, inv.l).norm().times(1/(float)1000);

        boolean good = p.move(v);
        double new_tile_score = tile.score();
        double new_goal_score = goal.score();
        double tile_delta = new_tile_score - tile_score;
        double goal_delta = 0;//new_goal_score - goal_score;
        double delta = tile_delta + goal_delta;
        //double swapProbability = Math.exp((-1 * delta) / temperature);
        //boolean doSwap = Math.random() < swapProbability;
        boolean doSwap = good && (tile_delta <= 0 && goal_delta <= 0);

        if (doSwap) {
            tile_score = new_tile_score;
            goal_score = new_goal_score;
            //System.out.println("score: " + tile_score + " / " + goal_score);
        } else {
            p.move(v.times(-1));
        }
    }

    public static int verts = 0;

    public void anneal() throws Exception {
        int verts = 0;
        while(true) {
            for(int i=0; i<1; i++) {
                repaint();
                for(Mesh.T t : tile)
                    for(Mesh.Vert p : new Mesh.Vert[] { t.v1(), t.v2(), t.v3() }) {
                        rand(10,p);
                    }
                goal.unscore();
                tile.unscore();
                goal.fundamental();
                tile.fundamental();
                goal.rescore();
                tile.rescore();
            }
            breakit();
       }
    }

    public static void main(String[] s) throws Exception {
        StlFile stlf = new StlFile();
        stlf.load("simplefish.stl");
        Frame f = new Frame();
        Main main = new Main(stlf, f);
        main.anneal();
    }
}